uniform sampler3D u_texWeight0;
uniform sampler3D u_texWeight1;
uniform bool u_flagNoise;
uniform bool u_flagLighting;
uniform float u_time;
varying vec3 v_position;
varying vec4 v_color;
varying vec3 v_normal;

extern float snoise(vec4 p);

void main() {
    // obtain weight parameters from a 3d texture according to the original color
    vec4 weight0 = texture3D(u_texWeight0, v_color.rgb);
    vec4 weight1 = texture3D(u_texWeight1, v_color.rgb);
    // evaluate noise at each frequency octave
    float base = 8.0;
    float n0 = u_flagNoise ? snoise(vec4(base *  1.0 * v_position, u_time)) : 0.0;
    float n1 = u_flagNoise ? snoise(vec4(base *  2.0 * v_position, u_time)) : 0.0;
    float n2 = u_flagNoise ? snoise(vec4(base *  4.0 * v_position, u_time)) : 0.0;
    float n3 = u_flagNoise ? snoise(vec4(base *  8.0 * v_position, u_time)) : 0.0;
    float n4 = u_flagNoise ? snoise(vec4(base * 16.0 * v_position, u_time)) : 0.0;
    float n5 = u_flagNoise ? snoise(vec4(base * 32.0 * v_position, u_time)) : 0.0;
    float n6 = u_flagNoise ? snoise(vec4(base * 64.0 * v_position, u_time)) : 0.0;
    float a  = weight0.r;
    float w0 = weight0.g;
    float w1 = weight0.b;
    float w2 = weight0.a;
    float w3 = weight1.r;
    float w4 = weight1.g;
    float w5 = weight1.b;
    float w6 = weight1.a;
    float n  = a * (
        w0 * n0 +
        w1 * n1 +
        w2 * n2 +
        w3 * n3 +
        w4 * n4 +
        w5 * n5 +
        w6 * n6) / (w0 + w1 + w2 + w3 + w4 + w5 + w6);
    float diffuse = pow(dot(normalize(v_normal), gl_LightSource[0].position.xyz) * 0.5 + 0.5, 0.75);        // half Lambert
    //float diffuse = dot(v_normal, gl_LightSource[0].position.xyz);        // standard Lambert
    diffuse = max(0.0, min(1.0, diffuse));
    if (!u_flagLighting)
        diffuse = 1.0;
    gl_FragColor.rgb = max(vec3(0.0), min(vec3(1.0), (v_color + vec3(n)) * diffuse));
    gl_FragColor.a = v_color.a;
}
